package ru.dbotthepony.mc.otm.worldgen.placement

import net.minecraft.core.BlockPos
import net.minecraft.util.RandomSource
import net.minecraft.util.valueproviders.FloatProvider
import net.minecraft.util.valueproviders.IntProvider
import ru.dbotthepony.mc.otm.core.math.component1
import ru.dbotthepony.mc.otm.core.math.component2
import ru.dbotthepony.mc.otm.core.math.component3
import ru.dbotthepony.mc.otm.core.math.plus
import java.util.stream.Stream

interface IEllipsoidPlacement {
	/**
	 * X position within ellipsoid sampler, expected to be in range -1 .. 1
	 */
	val x: FloatProvider

	/**
	 * Z position within ellipsoid sampler, expected to be in range -1 .. 1
	 */
	val z: FloatProvider

	/**
	 * Y position within ellipsoid sampler, expected to be in range -1 .. 1
	 */
	val y: FloatProvider

	/**
	 * Total amount of block positions to generate
	 */
	val count: IntProvider

	/**
	 * Ellipsoid size sampler on X axis
	 */
	val xLength: FloatProvider

	/**
	 * Ellipsoid size sampler on Z axis
	 */
	val zLength: FloatProvider

	/**
	 * Ellipsoid size sampler on Y axis
	 */
	val yLength: FloatProvider

	fun getEllipsoidPositions(random: RandomSource, position: BlockPos): Stream<BlockPos> {
		val count = count.sample(random)

		if (count <= 0)
			return Stream.empty()

		val xLength = xLength.sample(random)
		val zLength = zLength.sample(random)
		val yLength = yLength.sample(random)

		val xPow = xLength * xLength
		val zPow = zLength * zLength
		val yPow = yLength * yLength

		return Stream.generate {
			val x = this.x.sample(random) * xLength
			val y = this.y.sample(random) * yLength
			val z = this.z.sample(random) * zLength
			BlockPos(x.toInt(), y.toInt(), z.toInt())
		}
			.limit(count.toLong() * 10)
			.filter {
				val (ellipsoidX, ellipsoidY, ellipsoidZ) = it

				(ellipsoidX * ellipsoidX) / xPow +
					(ellipsoidY * ellipsoidY) / yPow +
					(ellipsoidZ * ellipsoidZ) / zPow <= 1.0f
			}
			.distinct()
			.limit(count.toLong())
			.map { it + position }
	}
}
